#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <iterator>

#include "plugin.hpp"

#include "types.hpp"
#include "types_xhrip.hpp"
#include "xHRiP.hpp"
#include "manager.hpp"
#include "tools.hpp"
#include "detector.hpp"
#include "lang.hpp"

PluginStartupInfo    startupInfo;
Options              op;
static HANDLE Settings;
Detector*            detector;

extern const GUID MainGuid = __uuidof(struct __declspec(uuid("{0DFA5BCB-4A11-42AB-AF24-0ADD6008B8B6}")) TMainGuid);

void WINAPI _export SetStartupInfoW(PluginStartupInfo *info)
{
  //    PluginStartupInfo
  startupInfo         = *info;

  FarSettingsCreate fsc{ sizeof(FarSettingsCreate), MainGuid, 0 };
  startupInfo.SettingsControl(INVALID_HANDLE_VALUE, SCTL_CREATE, PSL_LOCAL, &fsc);
  Settings = fsc.Handle;

  FarSettingsItem fsi{ sizeof(FarSettingsItem), 0 };

  fsi.Type = FST_QWORD;
  fsi.Name = L"ShowExt";
  fsi.Number = 1;

  startupInfo.SettingsControl(Settings, SCTL_GET, 0, &fsi);
  op.showExt = fsi.Number;

  fsi.Name = L"DefaultPanelMode";
  fsi.Number = 5;

  startupInfo.SettingsControl(Settings, SCTL_GET, 0, &fsi);
  op.defaultPanelMode = fsi.Number;

  fsi.Name = L"DefaultFormat";
  fsi.Number = 1;

  startupInfo.SettingsControl(Settings, SCTL_GET, 0, &fsi);
  op.defaultFormat = fsi.Number;

  fsi.Name = L"DetectFormat";
  fsi.Number = 1;

  startupInfo.SettingsControl(Settings, SCTL_GET, 0, &fsi);
  op.detectFormat = fsi.Number;

  fsi.Type = FST_STRING;
  fsi.Name = L"IniFilePath";
  fsi.String = L"";
  startupInfo.SettingsControl(Settings, SCTL_GET, 0, &fsi);
  wcsncpy(op.iniFilePath, fsi.String, std::size(op.iniFilePath));
  op.iniFilePath[std::size(op.iniFilePath) - 1] = 0;

  detector            = new Detector (op.iniFilePath, startupInfo.ModuleName);
}

void WINAPI _export ExitFARW()
{
  delete detector;
  detector = nullptr;

  startupInfo.SettingsControl(Settings, SCTL_FREE, 0, nullptr);
  Settings = nullptr;
}

HANDLE WINAPI _export AnalyseW(const struct AnalyseInfo *Info)
{
    const wchar_t *name{ Info->FileName };
    size_t dataSize{ Info->BufferSize };
    PUCHAR data{ (PUCHAR)Info->Buffer };
    if(name == NULL) return nullptr;

  //   HoBet'
  HoHdr hdr;
  CopyMemory(&hdr, data, sizeof(HoHdr));
  if(hdr.checkSum != calculateCheckSum(hdr)) return nullptr;

  char signature[] = { 'H', 'R', 'i' };
  char signature2[]= { 'H', 'r', 's', 't', '2' };
  int  offset      = sizeof(HoHdr);

  for(; offset < dataSize - 8 - 5; ++offset)
    if(!memcmp(data+offset, signature, sizeof(signature)) &&
       !memcmp(data+offset+8, signature2, sizeof(signature2)))
    {
      Manager *m = new Manager(name, offset);
      return (HANDLE)m;
    }

  return nullptr;
}

void WINAPI _export CloseAnalyseW(const struct CloseAnalyseInfo *Info)
{
    delete (Manager *)Info->Handle;
}

HANDLE WINAPI _export OpenW(const OpenInfo *Info)
{
    if(Info->OpenFrom == OPEN_ANALYSE)
    {
        OpenAnalyseInfo *An = (OpenAnalyseInfo *)Info->Data;
        return An->Handle;
    }
    if(Info->OpenFrom != OPEN_COMMANDLINE) return nullptr;

    const OpenCommandLineInfo *CommandLineInfo = (const OpenCommandLineInfo *)Info->Data;
    const wchar_t *name = CommandLineInfo->CommandLine;


  HANDLE hostFile = CreateFileW(name,
                               GENERIC_READ,
                               FILE_SHARE_READ | FILE_SHARE_WRITE,
                               NULL,
                               OPEN_EXISTING,
                               FILE_FLAG_SEQUENTIAL_SCAN,
                               NULL);
  if(hostFile == INVALID_HANDLE_VALUE) return nullptr;
  
  BYTE data[256*16+17];
  DWORD fileSize = GetFileSize(hostFile, NULL);
  if(fileSize < (sizeof(HoHdr) + 8 + 5))
  {
    CloseHandle(hostFile);
    return nullptr;
  }

  DWORD noBytesRead;
  fileSize = fileSize > sizeof(data) ? sizeof(data) : fileSize;
  ReadFile(hostFile, data, fileSize, &noBytesRead, NULL);
  CloseHandle(hostFile);

  HoHdr *hdr = (HoHdr*)data;
  if(hdr->checkSum != calculateCheckSum(*hdr)) return nullptr;

  char signature[] = { 'H', 'R', 'i' };
  char signature2[]= { 'H', 'r', 's', 't', '2' };
  int  offset      = sizeof(HoHdr);

  for(; offset < fileSize - 8 - 5; ++offset)
    if(!memcmp(data+offset, signature, sizeof(signature)) &&
       !memcmp(data+offset+8, signature2, sizeof(signature2)))
    {
      Manager *m = new Manager(name, offset);
      return (HANDLE)m;
    }

  return nullptr;
}

void WINAPI _export ClosePanelW(const ClosePanelInfo *Info)
{
    delete (Manager*)Info->hPanel;
}

intptr_t  WINAPI _export GetFindDataW(struct GetFindDataInfo *Info)
{
    Manager* m = (Manager *)Info->hPanel;
    return (m->getFindData(&Info->PanelItem, &Info->ItemsNumber, Info->OpMode));
}

void WINAPI _export FreeFindDataW(const struct FreeFindDataInfo *Info)
{
    Manager* m = (Manager *)Info->hPanel;
    m->freeFindData(Info->PanelItem, Info->ItemsNumber);
}

void WINAPI _export GetOpenPanelInfoW(struct OpenPanelInfo *Info)
{
    Manager* m = (Manager *)Info->hPanel;
    m->GetOpenPanelInfo(Info);
}

static const GUID MenuGuid = __uuidof(struct __declspec(uuid("{9B8FC386-E4B7-4D5E-A66A-002294DF9BE7}")) TMenuGuid);

void WINAPI _export GetPluginInfoW(PluginInfo *info)
{
    info->StructSize = sizeof(*info);
    info->Flags = PF_DISABLEPANELS;

    static const wchar_t *cmdPrefix = L"xHRiP";
    info->CommandPrefix = cmdPrefix;
    static const wchar_t *pluginCfgStrings[1];
    pluginCfgStrings[0] = GetMsg(MConfig);

    info->PluginMenu.Guids = &MenuGuid;
    info->PluginMenu.Strings = pluginCfgStrings;
    info->PluginMenu.Count = std::size(pluginCfgStrings);

    info->PluginConfig.Guids = &MenuGuid;
    info->PluginConfig.Strings = pluginCfgStrings;
    info->PluginConfig.Count = std::size(pluginCfgStrings);
}

intptr_t  WINAPI _export ProcessPanelEventW(const struct ProcessPanelEventInfo *Info)
{
  if(Info->Event == FE_CHANGEVIEWMODE)
  {
    startupInfo.PanelControl(Info->hPanel, FCTL_UPDATEPANEL, 1, nullptr);
  }
  if(Info->Event == FE_CLOSE)
  {
    PanelInfo panel = { sizeof(PanelInfo) };
    startupInfo.PanelControl(Info->hPanel, FCTL_GETPANELINFO, 0, &panel);

    FarSettingsItem fsi{ sizeof(FarSettingsItem), 0 };

    op.defaultPanelMode = panel.ViewMode;
    fsi.Type = FST_QWORD;
    fsi.Name = L"DefaultPanelMode";
    fsi.Number = panel.ViewMode;
    startupInfo.SettingsControl(Settings, SCTL_SET, 0, &fsi);
  }
  return FALSE;
}

intptr_t WINAPI _export GetFilesW(struct GetFilesInfo *Info)
{
    Manager* m = (Manager*)Info->hPanel;
    return m->getFiles(Info->PanelItem, Info->ItemsNumber, Info->Move, Info->DestPath, Info->OpMode);
}

intptr_t WINAPI _export ConfigureW(const struct ConfigureInfo *Info)
{
  FarDialogItem dlgItems[] =
  {
    { DI_DOUBLEBOX,   3,  1, 40, 13, {FALSE}, nullptr, nullptr,                                   0, GetMsg(MConfig) },
    { DI_CHECKBOX,    5,  2,  0,  0,  {TRUE}, nullptr, nullptr,                                   0, GetMsg(MDetectFormats) },
    { DI_CHECKBOX,    5,  3,  0,  0, {FALSE}, nullptr, nullptr,                                   0, GetMsg(MShowExt) },
    { DI_TEXT,        3,  4,  0,  0, {FALSE}, nullptr, nullptr,        DIF_BOXCOLOR | DIF_SEPARATOR, L"" },
    { DI_RADIOBUTTON, 5,  5,  0,  0, {FALSE}, nullptr, nullptr,                           DIF_GROUP, GetMsg(MHoBeta) },
    { DI_RADIOBUTTON, 5,  6,  0,  0, {FALSE}, nullptr, nullptr,                                   0, GetMsg(MSCL) },
    { DI_RADIOBUTTON, 5,  7,  0,  0, {FALSE}, nullptr, nullptr,                                   0, GetMsg(MAutoSelect) },
    { DI_TEXT,        3,  8,  0,  0, {FALSE}, nullptr, nullptr,        DIF_BOXCOLOR | DIF_SEPARATOR, L"" },
    { DI_TEXT,        5,  9,  0,  0, {FALSE}, nullptr, nullptr,                                   0, GetMsg(MIniPath) },
    { DI_EDIT,        5, 10, 38,  0, {FALSE}, nullptr, nullptr,                                   0, op.iniFilePath },
    { DI_TEXT,        3, 11,  0,  0, {FALSE}, nullptr, nullptr,        DIF_BOXCOLOR | DIF_SEPARATOR, L"" },
    { DI_BUTTON,      0, 12,  0,  0, {FALSE}, nullptr, nullptr, DIF_CENTERGROUP | DIF_DEFAULTBUTTON, GetMsg(MSave) },
    { DI_BUTTON,      0, 12,  0,  0, {FALSE}, nullptr, nullptr,                     DIF_CENTERGROUP, GetMsg(MCancel) }
  };

  dlgItems[1].Selected = op.detectFormat;
  dlgItems[2].Selected = op.showExt;
  int num = 3 + op.defaultFormat;
  if(num < 4 || num > 6) num = 6;
  dlgItems[num].Selected = TRUE;
  
  std::unique_ptr<void, decltype(SafeDlgFree) *> Dlg{ startupInfo.DialogInit(&MainGuid,
                                   &__uuidof(struct __declspec(uuid("{0707D133-FE46-4897-9CE3-014BDE280313}")) TDlgGuid),
                                   -1, -1, 44, 15,
                                   nullptr,
                                   dlgItems,
                                   std::size(dlgItems), 0, 0, startupInfo.DefDlgProc, nullptr), SafeDlgFree };
  if(Dlg.get() == INVALID_HANDLE_VALUE)
  {
      return FALSE;
  }
  intptr_t askCode = startupInfo.DialogRun(Dlg.get());
  if(askCode != 11) return FALSE;
  
  op.detectFormat = startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 1, 0);
  op.showExt = startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 2, 0);

  op.defaultFormat = 3;
  if(startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 4, 0)) op.defaultFormat = 1;
  if(startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 5, 0)) op.defaultFormat = 2;

  FarDialogItemData ItemData = { sizeof(FarDialogItemData), std::size(op.iniFilePath) - 1, op.iniFilePath };
  startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 9, &ItemData);

  FarSettingsItem fsi{ sizeof(FarSettingsItem), 0 };
  fsi.Type = FST_QWORD;

  fsi.Name = L"DetectFormat";
  fsi.Number = op.detectFormat;
  startupInfo.SettingsControl(Settings, SCTL_SET, 0, &fsi);

  fsi.Name = L"ShowExt";
  fsi.Number = op.showExt;
  startupInfo.SettingsControl(Settings, SCTL_SET, 0, &fsi);

  fsi.Name = L"DefaultFormat";
  fsi.Number = op.defaultFormat;
  startupInfo.SettingsControl(Settings, SCTL_SET, 0, &fsi);

  fsi.Type = FST_STRING;

  fsi.Name = L"IniFilePath";
  fsi.String = op.iniFilePath;
  startupInfo.SettingsControl(Settings, SCTL_SET, 0, &fsi);

  return TRUE;
}

void WINAPI GetGlobalInfoW(struct GlobalInfo *Info)
{
    Info->StructSize = sizeof(*Info);
    Info->MinFarVersion = MAKEFARVERSION(3, 0, 0, 2927, VS_RELEASE);   //   Far Manager,    API
    Info->Version = MAKEFARVERSION(1, 0, 0, 1, VS_RC);   //   1.0.0.1, -
    Info->Guid = MainGuid;
    Info->Title = L"xHRiP";
    Info->Description = L"xHRiP";
    Info->Author = L"Elf/2 & Alexandr Medvedev";
}
