#ifndef manager_hpp
#define manager_hpp

#include "plugin.hpp"
#include "FmtPlugin.hpp"
#include "iSDOS.hpp"

class Manager
{
  public:
    Manager(const wchar_t* fileName, FmtPlugin* fmtPlugin);
    ~Manager();

    void GetOpenPanelInfo(struct OpenPanelInfo *info);
    int  getFindData      (PluginPanelItem **pPanelItem, size_t *pNoItems, int opMode);
    void freeFindData     (PluginPanelItem *panelItem,   int  noItems);
    int  setDirectory     (const wchar_t* dirName, int opMode);

    bool openHostFile     (void);
    bool closeHostFile    (void);
    bool readBlock        (u16 num, u8* buf);
    bool writeBlock       (u16 num, u8* buf);
    
  private:
    FmtPlugin* filer;
    HANDLE     img;
    
    wchar_t    hostFileName[300];
    wchar_t    curDir[300];

    u16        noFreeBlocks;
    UniHdr     device_sys_hdr;
    u8*        device_sys{};
    
    DiskHdr    dsk;
    UniHdr     files[128];
    wchar_t    pcFileName[128][8+3+2];

    WIN32_FIND_DATAW lastModifed;

  private:
    bool readFolder      (const UniHdr& h, UniHdr* dest);
    bool readFolder      (u16 firstBlock,  UniHdr* dest);
    bool writeFolder     (UniHdr* pDir);
    bool readInfo        (void);
    u16  getNoFreeBlocks (void);
    void read_device_sys (void);
    void write_device_sys(void);
    
  public:
    int writeFile (const wchar_t *name, const UniHdr& hdr);
    int getOneFile(UniHdr& h, UniHdr* pDir, const wchar_t* to_path, const wchar_t* from_path, int move, Action& action, int opMode);
    int getFiles  (PluginPanelItem *panelItem, int noItems, int move, const wchar_t *destPath, int opMode);

  private:
    bool   isFreeBlock      (u16 n);
    bool   markBlock        (u16 n);
    bool   unmarkBlock      (u16 n);
    int    findFreeBlocks   (u16 n);
    bool   findMaxFreeBlocks(u16 n, u16* firstFoundBlock, u16* noFoundBlocks);

  public:
    int deleteOneFile(UniHdr& h, UniHdr* dir, const wchar_t* dirName, int opMode, Action& actProtected, Action& actFolder);
    int deleteFiles(PluginPanelItem *panelItem, int noItems, int opMode);

  private:
    bool reserveSpaceForFiles(UniHdr* pDir, u16 noAddFiles);
    int  makeFolder          (UniHdr* pDir, const wchar_t* name);
  public:
    int  makeDirectory(const wchar_t *dirName, int opMode);
    
  private:
    int putOneFile(UniHdr* pDir, const WIN32_FIND_DATAW &file, const wchar_t* fromDir, const wchar_t* dirName, int move, Action& action, int opMode);
    int putOneFolder(UniHdr* pDir, const wchar_t *FileName, const wchar_t* fromDir, const wchar_t* dirName, int move, Action& action, int opMode);
  public:
    int putFiles  (PluginPanelItem *panelItem, int noItems, int move, int opMode);
    int processKey(const INPUT_RECORD *Rec);
};

#endif
