#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#include "plugin.hpp"

#include "types.hpp"
#include "types_isd.hpp"
#include "manager.hpp"
#include "tools.hpp"
#include "tools_isd.hpp"
#include "iSDOS.hpp"
#include "iSDOS_tools.hpp"
#include "iterator.hpp"
#include "lang.hpp"

extern PluginStartupInfo startupInfo;
extern Options           op;

int Manager::deleteOneFile(UniHdr& h, UniHdr* pDir, const wchar_t* dirName, int opMode, Action& actProtected, Action& actFolder)
{
  wchar_t name[8+3+2];
  makeName(h, name);

  wchar_t fullName[300];
  makeFullName(fullName, dirName, name);

  wchar_t shortName[40];
  cropLongName(shortName, fullName, 39);
  
  int result = 1;
  if(isDir(h))
  {
    if(actFolder == ASK_USER)
    {
      const wchar_t *msgItems[7];
      msgItems[0] = GetMsg(MDeleteFolder);
      msgItems[1] = GetMsg(MDelFolder);
      msgItems[2] = shortName;
      msgItems[3] = GetMsg(MDeleteButton);
      msgItems[4] = GetMsg(MAll);
      msgItems[5] = GetMsg(MSkip);
      msgItems[6] = GetMsg(MCancel);
      int askCode = messageBox(FMSG_WARNING /*| FMSG_DOWN*/,
                               msgItems,
                               sizeof(msgItems)/sizeof(msgItems[0]), 4);
      switch(askCode)
      {
        case -1:
        case 3:
                  return -1;
        case 2:
                  return 0;
        case 1:
                  actFolder = DO_ALL;
      }
    }
    UniHdr curDir[128];
    readFolder(h, curDir);
    
    for(int i = 1; i < curDir[0].dir.totalFiles; ++i)
      if(curDir[i].attr & FLAG_EXIST)
      {
        
        int tmp_result = deleteOneFile(curDir[i], curDir, fullName, opMode, actProtected, actFolder);
        if(tmp_result == 1) continue;
        
        result = tmp_result;
        if(result == -1) break;
      }
    if(result == 1)
    {
      h.attr &= ~FLAG_EXIST;
      curDir[0].attr &= ~FLAG_EXIST;
      Iterator cur(filer, img, h), end;
      while(cur != end)
      {
        unmarkBlock(*cur);
        ++cur;
      }
      if(!(h.attr & FLAG_SOLID)) unmarkBlock(h.dir.descr1stBlock);
      
      --pDir[0].dir.noFiles;
    }
    writeFolder(curDir);
  }
  else
  {
    if(h.attr == 0xFF && h.file.systemFlag == 0xFF) return 0;
    if(h.attr & FLAG_DELETE_PROTECT)
    {
      if(actProtected == SKIP_ALL)  return 0;
      if((opMode & OPM_SILENT) == 0 && actProtected == ASK_USER)
      {
        const wchar_t *msgItems[9];
        msgItems[0] = GetMsg(MWarning);
        msgItems[1] = GetMsg(MFileProtected);
        msgItems[2] = shortName;
        msgItems[3] = GetMsg(MDelProtected);
        msgItems[4] = GetMsg(MDeleteButton);
        msgItems[5] = GetMsg(MAll);
        msgItems[6] = GetMsg(MSkip);
        msgItems[7] = GetMsg(MSkipAll);
        msgItems[8] = GetMsg(MCancel);
        int askCode = messageBox(FMSG_WARNING /*| FMSG_DOWN*/,
                                 msgItems,
                                 sizeof(msgItems)/sizeof(msgItems[0]), 5);
        switch(askCode)
        {
          case -1:
          case 4:
                    return -1;
          case 3:
                    actProtected = SKIP_ALL;
          case 2:
                    return 0;
          case 1:
                    actProtected = DO_ALL;
        }
      }
    }
    if((opMode & OPM_SILENT) == 0)
    {
      const wchar_t *msgItems[3];
      msgItems[0] = GetMsg(MDelete);
      msgItems[1] = GetMsg(MDeleting);
      msgItems[2] = shortName;
      messageBox(0, msgItems, sizeof(msgItems)/sizeof(msgItems[0]), 0);
    }
    --pDir[0].dir.noFiles;
    h.attr &= ~FLAG_EXIST;

    Iterator cur(filer, img, h), end;
    while(cur != end)
    {
      unmarkBlock(*cur);
      ++cur;
    }
    if(!(h.attr & FLAG_SOLID)) unmarkBlock(h.file.firstBlock);
  }
  return result;
}

int Manager::deleteFiles(PluginPanelItem *panelItem, int noItems, int opMode)
{
  if(noItems == 0) return TRUE;
  readInfo();

  if((opMode & OPM_SILENT) == 0)
  {
    const wchar_t *msgItems[5];
    msgItems[0] = GetMsg(MDelete);
    msgItems[1] = GetMsg(MDelFile);
    msgItems[3] = GetMsg(MDeleteButton);
    msgItems[4] = GetMsg(MCancel);

    wchar_t msg[30];
    if(noItems == 1)
      wsprintfW(msg, L"%s", panelItem[0].FileName);
    else
      wsprintfW(msg, GetMsg(MDelFiles), noItems);

    msgItems[2] = msg;
    if(messageBox(0, msgItems, sizeof(msgItems)/sizeof(msgItems[0]), 2)!=0) return FALSE;
    if(noItems > 1)
    {
      const wchar_t *msgItems[5];
      msgItems[0] = GetMsg(MDeleteFiles);
      msgItems[1] = GetMsg(MDelFile);
      msgItems[3] = GetMsg(MAll);
      msgItems[4] = GetMsg(MCancel);

      wchar_t msg[30];
      wsprintfW(msg, GetMsg(MDelFiles), noItems);
      msgItems[2] = msg;
      if(messageBox(FMSG_WARNING, msgItems, sizeof(msgItems)/sizeof(msgItems[0]), 2)!=0)
        return FALSE;
    }
  }
  if(!openHostFile()) return FALSE;

  HANDLE screen = startupInfo.SaveScreen(0, 0, -1, -1);

  Action  actProtected = ASK_USER;
  Action  actFolder    = ASK_USER;
  int     exitCode     = TRUE;
  
  for(int iNum = 0; iNum < noItems; ++iNum)
  {
    int fNum = findFile(files, panelItem[iNum].FileName);
    if(!fNum) continue;

    int result = deleteOneFile(files[fNum], files, curDir, opMode, actProtected, actFolder);

    if(result == -1)
    {
      exitCode = FALSE;
      break;
    }
  }

  startupInfo.RestoreScreen(screen);

  writeFolder(files);
  write_device_sys();
  closeHostFile();
  op.reread = true;
  startupInfo.PanelControl(PANEL_PASSIVE, FCTL_UPDATEPANEL, 1, nullptr);
  op.reread = false;
  return exitCode;
}
