#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <iterator>

#include "plugin.hpp"

#include "types.hpp"
#include "detector_xlook.hpp"
#include "tools.hpp"

extern PluginStartupInfo startupInfo;

namespace xlook
{

void DetectorManager::LoadFolder(const wchar_t *path)
{
    wchar_t plugin_mask[300];

    lstrcpyW(plugin_mask, path);
    lstrcatW(plugin_mask, L"*.dtr");

    WIN32_FIND_DATAW data;
    HANDLE foundFile = FindFirstFileW(plugin_mask, &data);

    int done = (foundFile == INVALID_HANDLE_VALUE);

    while(!done)
    {
        wchar_t plugin_name[300];
        lstrcpyW(plugin_name, path);
        lstrcatW(plugin_name, data.cFileName);

        HMODULE h = LoadLibraryW(plugin_name);
        if(h)
        {
            Detector& cur = detector[no_detectors];

            cur.handle = h;
            cur.init = (init_t)GetProcAddress(h, "init");
            cur.detect = (detect_t)GetProcAddress(h, "detect");
            cur.extract = (extract_t)GetProcAddress(h, "extract");
            cur.cleanup = (cleanup_t)GetProcAddress(h, "cleanup");

            if(!cur.detect)
                FreeLibrary(h);
            else
            {
                if(cur.init) cur.init(path);
                ++no_detectors;
            }
            if(no_detectors == 256) break;
        }
        done = !FindNextFileW(foundFile, &data);
    }
    FindClose(foundFile);

    lstrcpyW(plugin_mask, path);
    lstrcatW(plugin_mask, L"*.*");

    foundFile = FindFirstFileW(plugin_mask, &data);

    done = (foundFile == INVALID_HANDLE_VALUE);

    while(!done)
    {
        if((data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) &&
            lstrcmpW(data.cFileName, L".") &&
            lstrcmpW(data.cFileName, L".."))
        {
            wchar_t folder_name[300];
            lstrcpyW(folder_name, path);
            lstrcatW(folder_name, data.cFileName);
            lstrcatW(folder_name, L"\\");

            LoadFolder(folder_name);
        }

        done = !FindNextFileW(foundFile, &data);
    }
    FindClose(foundFile);
}

DetectorManager::DetectorManager()
{
    no_detectors = 0;

    wchar_t plugin_path[300];

    lstrcpyW(plugin_path, startupInfo.ModuleName);
    lstrcpyW(pointToName(plugin_path), L"detectors\\");

    LoadFolder(plugin_path);

    if(!no_detectors)
    {
        const wchar_t *msgItems[] = { L"xLook Warning", L"No detectors loaded", L"Ok" };
        messageBox(FMSG_WARNING, msgItems, std::size(msgItems), 1);
    }
}

DetectorManager::~DetectorManager()
{
    for(int i = 0; i < no_detectors; ++i)
    {
        Detector& cur = detector[no_detectors];

        if(cur.cleanup) cur.cleanup();
        FreeLibrary(cur.handle);
    }
}

Detector* DetectorManager::find(DetectorResult &result, const wchar_t *name, const unsigned char *data, int file_size)
{
    for(int i = 0; i < no_detectors; ++i)
        if(detector[i].detect(result, name, (HoHdr*)data, data + sizeof(HoHdr), file_size - sizeof(HoHdr))) return &detector[i];

    return 0;
}
}
