#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#include "../../dtr.hpp"

//   mspack  hrust
static BYTE* base;
static BYTE* p;
static int   idx;
//int   len;
//bool  eof;
static WORD  bits;

static BYTE getByte(void)
{
    //if(p - base == len) { eof = true; return 0; }
    return *p++;
}

static BYTE getBit()
{
    WORD mask[]  = { 0x8000, 0x4000, 0x2000, 0x1000, 0x800, 0x400, 0x200, 0x100, 0x80, 0x40, 0x20, 0x10, 0x8, 0x4, 0x2, 0x1 };

    BYTE bit = (bits & mask[idx]) ? 1 : 0;
    if(idx == 15)
    {
        bits  = getByte();
        bits += 256*getByte();
    }

    idx = (idx + 1) % 16;
    return bit;
}

static BYTE getBits(int n)
{
    BYTE r = 0;
    do { r = 2*r + getBit(); } while(--n);
    return r;
}

// mspack
static WORD calculateDisp(void)
{
    if(getBit()) return getByte() + 1;

    BYTE code = getBits(3);
    if(code < 2) return 256*(code+1) + getByte() + 1;

    code = 2*code + getBit();
    if(code < 8) return 256*(code-1) + getByte() + 1;

    code = 2*code + getBit();
    if(code < 0x17) return 256*(code-9) + getByte() + 1;

    code = 2*code + getBit();
    if((code&0x1f) < 0x1f)return 256*(code&0x1f) + getByte() + 1;

    WORD disp = 256*getByte();
    disp += getByte() + 1;
    return disp;
}

// mspack
static bool unmspk(BYTE *source, BYTE *unpackBuffer, DWORD *noUnpackedBytes)
{
    BYTE *to = unpackBuffer;
    BYTE *from = source + 0xF3;
    base = p = from;
    
    idx = 0;
    bits  = getByte();
    bits += 256*getByte();

//  if(*(WORD*)(source + 0xE9) - *(WORD*)(source + 0xEB) !=5)
//      packed_size = *(WORD*)(source + 0xED) + *(WORD*)(source + 0xEF) - *(WORD*)(source + 0xF1);
//  else
//      packed_size = *(WORD*)(source + 0xED) - *(WORD*)(source + 0xF1) + 1;

    while(1)
    {
        while(!getBit()) *to++ = getByte();

        WORD len = 0;
        BYTE bb  = 0;

        do
        {
            bb = getBits(2);
            len += bb;
        } while(bb == 0x03 && len != 0x0f);

        len += 2;

        if(len == 2)
        {
            WORD disp = getByte() + 1;
            for(BYTE i = 0; i < 2; ++i) *to++ = to[-disp];
            continue;
        }

        if(len == 5)
        {
            BYTE b = getByte();
            if(b == 0xff) break; // EOF
            if(b == 0xfe)
                { len = getByte(); len += 256*getByte(); }
            else
                len = b + 17;
        }
        else
            if(len > 5) --len;

        WORD disp = calculateDisp();

        for(WORD i = 0; i < len; ++i)
            *to++ = to[-disp];
    }
    for(BYTE i = 0; i < 5; ++i)
        *to++ = getByte();

    *noUnpackedBytes = to-unpackBuffer;

    return true;
}


// hrust 1
static bool unhr1(BYTE *source, BYTE *unpackBuffer, DWORD *noUnpackedBytes)
{
    BYTE *to = unpackBuffer;
    BYTE *from = source+12;
    base = p = from;
    
    //len = from[4] + 256*from[5];//blockSize
    idx = 0;

    bits  = getByte();
    bits += 256*getByte();

    *to++ = getByte();

    BYTE noBits = 2;
    BYTE mask[] = { 0, 0, 0xfc, 0xf8, 0xf0, 0xe0, 0xc0, 0x80, 0 };

    while(1)
    {
        while(getBit()) *to++ = getByte();

        WORD len = 0;
        BYTE bb  = 0;
        do
        {
            bb = getBits(2);
            len += bb;
        } while(bb == 0x03 && len != 0x0f);

        short offset = 0;

        if(len == 0)
        {
            offset = 0xfff8 + getBits(3);
            *to++ = to[offset];
            continue;
        }

        if(len == 1)
        {
            BYTE code = getBits(2);

            if(code == 2)
            {
                BYTE b = getByte();
                if(b >= 0xe0)
                {
                    b <<= 1; ++b; // rlca
                    b ^= 2;       // xor c

                    if(b == 0xff) { ++noBits; continue; }

                    offset = 0xff00 + b - 0x0f;

                    *to++ = to[offset];
                    *to++ = getByte();
                    *to++ = to[offset];
                    continue;
                }
                offset = 0xff00 + b;
            }

            if(code == 0 || code == 1)
            {
                offset = getByte();
                offset += 256*(code ? 0xfe : 0xfd);
            }
            if(code == 3) offset = 0xffe0 + getBits(5);

            for(BYTE i = 0; i < 2; ++i)
                *to++ = to[offset];
            continue;
        }

        if(len == 3)
        {
            if(getBit())
            {
                offset = 0xfff0 + getBits(4);
                *to++ = to[offset];
                *to++ = getByte();
                *to++ = to[offset];
                continue;
            }

            if(getBit())
            {
                BYTE noBytes = 6 + getBits(4);
                for(BYTE i = 0; i < 2*noBytes; ++i)
                    *to++ = getByte();
                continue;
            }

            len = getBits(7);
            if(len == 0x0f) break; // EOF
            if(len <  0x0f) len = 256*len + getByte();
        }

        if(len == 2) ++len;

        BYTE code = getBits(2);

        if(code == 1)
        {
            BYTE b = getByte();

            if(b >= 0xe0)
            {
                if(len > 3) return false;

                b <<= 1; ++b; // rlca
                b ^= 3;       // xor c

                offset = 0xff00 + b - 0x0f;

                *to++ = to[offset];
                *to++ = getByte();
                *to++ = to[offset];
                continue;
            }
            offset = 0xff00 + b;
        }

        if(code == 0) offset = 0xfe00 + getByte();
        if(code == 2) offset = 0xffe0 + getBits(5);
        if(code == 3)
        {
            offset  = 256*(mask[noBits] + getBits(noBits));
            offset += getByte();
        }

        for(WORD i = 0; i < len; ++i)
            *to++ = to[offset];
    }

    for(int i = 0; i < 6; ++i)
        *to++ = source[6+i];

    *noUnpackedBytes = to-unpackBuffer;

    return true;
}

    
static bool unlz(BYTE *from, BYTE *unpackBuffer, DWORD *noUnpackedBytes)
{
    BYTE *to = unpackBuffer;

    while(*from != 0x80)
    {
        BYTE b = *from++;
        BYTE i;
        switch(b & 0xc0)
        {
        case 0x80: // copy
            for(i = 0; i < (b & 0x3f); ++i) *to++ = *from++;
            break;
        case 0xc0: // RLE
            for(i = 0; i < (b & 0x3f)+3; ++i) *to++ = *from;
            ++from;
            break;
        default: // backreference (LZ)
            {
                BYTE len    = ((b & 0xf0)>>4) + 3;
                WORD offset = (256*(b&0x0f)) + *from++;
                for(BYTE i = 0; i < len; ++i) *to++ = *(to-offset);
            }
        }
    }

    *noUnpackedBytes = to-unpackBuffer;

    return true;
}

static bool uncc3(BYTE *from, BYTE *unpackBuffer, DWORD *noUnpackedBytes)
{
    BYTE *to   = unpackBuffer;

    while((*from & 0x0f) != 0x03)
    {
        BYTE b = *from++;
        if((b & 0x01) == 0)  // backreference
        {
            BYTE len    = ((b & 0xf0)>>4) + 3;
            WORD offset = (128*(b&0x0f)) + *from++;

            for(BYTE i = 0; i < len; ++i) *to++ = to[-offset];

            if((b & 0xf0) == 0xf0)
            {
                BYTE len2 = *from++;
                if(len2 != 0)
                    for(BYTE i = 0; i < len2; ++i) *to++ = to[-offset];
            }
            continue;
        }

        switch(b & 0x0f)
        {
        case 0x09: // short RLE [1001]
            {
                BYTE len = ((b & 0xf0)>>4) + 3;
                for(BYTE i = 0; i < len; ++i) *to++ = *from;
                ++from;
                break;
            }
        case 0x01: // long RLE [0001]
            {
                WORD len = 256*((b&0xf0)>>4) + (*from++)+3;
                for(WORD i = 0; i < len; ++i) *to++ = *from;
                ++from;
                break;
            }
        case 0x0B: // 2 bytes (00-0x0E,0xFF) [1011] {0011 - end}
            {
                BYTE bb = ((b&0xf0)>>4) - 1;
                *to++ = bb;
                *to++ = bb;
                break;
            }
        case 0x05: // short copy [0101]
            {
                BYTE len = ((b & 0xf0)>>4) + 1;
                for(BYTE i = 0; i < len; ++i) *to++ = *from++;
                break;
            }
        case 0x0d: // long copy [1101]
            {
                WORD len = 256*((b & 0xf0)>>4) + (*from++) + 1;
                for(WORD i = 0; i < len; ++i) *to++ = *from++;
                break;
            }
        default: // short backreference for 2 bytes [x111]
            {
                WORD offset = (b & 0xf8)>>3;
                *to++ = *(to-offset);
                *to++ = *(to-offset);
            }
        }
    }

    *noUnpackedBytes = to-unpackBuffer;

    return true;
}

static void deleteColorControls(BYTE *unpackBuffer, DWORD *noUnpackedBytes)
{
    for(int i=0; i<(*noUnpackedBytes)-3; i++)
    {
        if(unpackBuffer[i] == 0x01 || unpackBuffer[i] == 0x02)
        {
            CopyMemory(unpackBuffer+i, unpackBuffer+i+1, (*noUnpackedBytes)-i-1);
            *noUnpackedBytes-=1;
            i--;
        }
        if(unpackBuffer[i] == 16)
        {
            CopyMemory(unpackBuffer+i, unpackBuffer+i+2, (*noUnpackedBytes)-i-2);
            *noUnpackedBytes-=2;
            i--;
        }
        if(unpackBuffer[i] == 24)
        {
            //unpackBuffer[i]='<';unpackBuffer[i+1]='?';unpackBuffer[i+2]='>';
            CopyMemory(unpackBuffer+i, unpackBuffer+i+3, (*noUnpackedBytes)-i-3);
            *noUnpackedBytes-=3;
            i--;
        }
    }
}

bool WINAPI extract(HoHdr *hdr, const BYTE *source, DWORD sourceSize, BYTE *unpackBuffer, DWORD *noUnpackedBytes, DWORD internal_id)
{
    switch(internal_id>>16)
    {
    case 0x01:
        unlz((BYTE*)source+(internal_id&0xFFFF), unpackBuffer, noUnpackedBytes);
        break;
    case 0x02:
        unhr1((BYTE*)source+(internal_id&0xFFFF), unpackBuffer, noUnpackedBytes);
        deleteColorControls(unpackBuffer, noUnpackedBytes);
        break;
    case 0x03:
        uncc3((BYTE*)source+(internal_id&0xFFFF), unpackBuffer, noUnpackedBytes);
        break;
    case 0x04:
        unmspk((BYTE*)source+(internal_id&0xFFFF), unpackBuffer, noUnpackedBytes);
        deleteColorControls(unpackBuffer, noUnpackedBytes);
        //*noUnpackedBytes = sourceSize-(internal_id&0xFFFF);
        //CopyMemory(unpackBuffer, source+(internal_id&0xFFFF), *noUnpackedBytes);
        break;
    //case 0x0x://lmg
        //unhr1((BYTE*)source+(internal_id&0xFFFF), unpackBuffer, noUnpackedBytes);
        //*noUnpackedBytes-=0x0F82;
        //CopyMemory(unpackBuffer, unpackBuffer+0x0F82, *noUnpackedBytes);
        //break;
    //case 0x0x://hrust2
    //  unHR2(source+(internal_id&0xFFFF), unpackBuffer, noUnpackedBytes);
    //  deleteACEControls(unpackBuffer);
    //  break;
    default:
        return false;
    }

    return true;
}

static bool checkFormat(BYTE *data, int offs, BYTE* sign, int signSize, int signOffs, int SFXsize, DWORD *id, BYTE pack)
{
    for(int i = 0; i < signSize; ++i)
        if(data[offs+signOffs+i] != sign[i])
            return false;

    *id=(pack<<16)+offs+SFXsize;
    return true;
}

bool WINAPI detect(DetectorResult &result, const wchar_t*, const HoHdr *hdr, const BYTE *data, int file_size)
{
    if(hdr->type != 'B')
        return false;

    //result.internal_id  pack_method<<16 + packed_data_offset
    //pack_method:
    //01 - LZS
    //02 - HRUST 1.x
    //03 - CC3
    //04 - MSPACK

    while(true)
    {
//////////  ZX WORD Text viewer by OLDMAN   by KSA
        BYTE LZSsignature[] = {0x06, 0, 0x7E, 0xCB, 0x7F, 0x20, 0x1D, 0xE6, 0x0F, 0x47};

        if(checkFormat((BYTE*)data, 0x0E00, LZSsignature, sizeof(LZSsignature), 0x26, 0x82, &result.internal_id, 0x01))
        {
            lstrcpyW(result.title, L"[ZX-WORD TEXT VIEWER v2.4 by OLDMAN]");
            break;
        }

        if(checkFormat((BYTE*)data, 0x0F00, LZSsignature, sizeof(LZSsignature), 0x26, 0x82, &result.internal_id, 0x01))
        {
            lstrcpyW(result.title, L"[ZX-WORD TEXT VIEWER v2.0 by OLDMAN]");
            break;
        }

        if(checkFormat((BYTE*)data, 0x1000, LZSsignature, sizeof(LZSsignature), 0x26, 0x82, &result.internal_id, 0x01))
        {
            lstrcpyW(result.title, L"[ZX-WORD TEXT VIEWER v2.2, v2.3 by OLDMAN]");
            break;
        }

        if(checkFormat((BYTE*)data, 0x1200, LZSsignature, sizeof(LZSsignature), 0x26, 0x82, &result.internal_id, 0x01))
        {
            lstrcpyW(result.title, L"[ZX-WORD TEXT VIEWER v1.1 by OLDMAN]");
            break;
        }

        if(checkFormat((BYTE*)data, 0x0C00, LZSsignature, sizeof(LZSsignature), 0x26, 0x82, &result.internal_id, 0x01))
        {
            lstrcpyW(result.title, L"[ZX-WORD TEXT VIEWER v1.3 by KSA SoftWare]");
            break;
        }

        if(file_size>0xC00)
        {
            if(data[0xC00] == 0xF3 && data[0xC01] == 'K' && data[0xC02] == 'S' && data[0xC03] == 'A' && data[0xC04] == 0x21)
            {
                result.internal_id=(0x03<<16)+0x0C99;
                lstrcpyW(result.title, L"[ZX-WORD TEXT VIEWER v1.5 by KSA SoftWare]");
                break;
            }
        }
////////TextViwer v2.01 by Eraser
        if(file_size>0x1400)
        {
            if(data[0x1503] == 'H' && data[0x1504] == 'R')
            {
                result.internal_id=(0x02<<16)+0x1503;
                lstrcpyW(result.title, L"[TextViwer v2.01 by Eraser'Delirium Tremens]");
                break;
            }
        }
////////Fastzero text viewer by RAGE
        if(file_size>0x1F00 && data[0x9F]=='T' && data[0xA0]=='E' && data[0xA1]=='X' && data[0xA2]=='T')
        {
            WORD offset=(0x1F+data[0x9E]+data[0xAB]+1)<<8;
            if(data[offset+0xE5] == 'M' && data[offset+0xE6] == 's' && data[offset+0xE7] == 'P' && data[offset+0xE8] == 'k')
            {
                result.internal_id=(0x04<<16)+offset;
                lstrcpyW(result.title, L"[Fastzero text viewer v1.0 by RAGE]");
                break;
            }
        }

        if(file_size>0x2200 && data[0xB8]=='T' && data[0xB9]=='E' && data[0xBA]=='X' && data[0xBB]=='T')
        {
            WORD offset=(0x22+data[0xAB]+data[0xB4])<<8;
            if(data[offset+0xE5] == 'M' && data[offset+0xE6] == 's' && data[offset+0xE7] == 'P' && data[offset+0xE8] == 'k')
            {
                result.internal_id=(0x04<<16)+offset;
                lstrcpyW(result.title, L"[Fastzero text viewer v1.2 by RAGE]");
                break;
            }
        }
////////TextViwer v3.6 by Spectre
        /*  if(file_size>0x100)
        {
            if(data[0xB8] == 'S' && data[0xB9] == 'p' && data[0xBA] == 'e' && data[0xBB] == 'c' && data[0xBC] == 't' && data[0xBD] == 'r' && data[0xBE] == 'e')
            {
                if(data[0xC7]=='7')
                {
                    result.internal_id=(0x04<<16)+0x1800+256*data[0x5F]+256*data[0x67];
                    lstrcpy(result.title, "[ZX-WORD TEXT VIEWER v1.4 by A.NIKIFOROV]");
                }
                else
                {
                    result.internal_id=(0x04<<16)+0x1D00+256*data[0x5F]+256*data[0x67];
                    lstrcpy(result.title, "[Spectre TextViever v3.6]");
                }
                break;
            }
        }

        /*if(file_size>0x100)
        {
            if(data[0x0203] == 'H' && data[0x0204] == 'R')
            {
                result.internal_id=(0x0x<<16)+0x203;
                lstrcpy(result.title, "[Lamergy]");
                break;
            }
        }*/

        /*if(file_size>0x1400)
        {
            int offs=0x1100;

            // 
            if(data[offs] != 'h' || data[offs+1] != 'r' || data[offs+2] != '2') return false;

            //  - 
            if((WORD*)*(offs+8)==0x1B00)
                offs+=(WORD*)*(offs+6);

            if(data[offs] != 'h' || data[offs+1] != 'r' || data[offs+2] != '2') return false;

            //  - 
            BYTE tmpBuf[0x10000]
            unHR2(data+offs, tmpBuf);
            if(!lstrcmp(tmpBuf, "ProTracker3")
                offs+=(WORD*)*(offs+6);
            if(!lstrcmp(tmpBuf, "ProTracker2")
                offs+=(WORD*)*(offs+6);

            if(data[offs] != 'h' || data[offs+1] != 'r' || data[offs+2] != '2') return false;

            result.internal_id=(0x0x<<16)+offs;
            lstrcpy(result.title, "[NViewer 1.0 by Spectre]");
            break;
        }*/

        return false;
    }

    result.action = do_extract;
    //lstrcpy(result.title, "[text viewer]");
    lstrcpyW(result.ext, L".txt");
    return true;

}
