#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <algorithm>
#include "plugin.hpp"

#include "types.hpp"
#include "types_scl.hpp"
#include "manager.hpp"
#include "detector.hpp"
#include "tools.hpp"
#include "lang.hpp"

extern PluginStartupInfo startupInfo;
extern Detector*         detector;
extern Options           op;

Manager::Manager(const wchar_t* fileName)
{
  lstrcpyW(hostFileName, fileName);
  
  noFiles = 0;
  
  ZeroMemory(&lastModifed, sizeof(WIN32_FIND_DATA));
}

Manager::~Manager()
{
}

void Manager::GetOpenPanelInfo(struct OpenPanelInfo *Info)
{
  Info->StructSize    = sizeof(*Info);
  Info->Flags         = /*OPIF_USEFILTER | OPIF_USEHIGHLIGHTING |*/ OPIF_ADDDOTS | OPIF_SHOWNAMESONLY | OPIF_SHOWPRESERVECASE;
  Info->HostFile      = hostFileName;
  Info->CurDir        = L"";
  Info->Format        = L"SCL";
  Info->StartSortMode = SM_UNSORTED;
  
  static wchar_t panelTitle[260];
  panelTitle[0] = 0;
  lstrcatW(panelTitle, L" SCL:");
  lstrcatW(panelTitle, pointToName(hostFileName));
  lstrcatW(panelTitle, L" ");
  Info->PanelTitle       = panelTitle;

  static PanelMode mode[10];
  static const wchar_t*     columnTitles3[4];
  columnTitles3[0] = GetMsg(MName);
  columnTitles3[1] = GetMsg(MSectorSize);
  columnTitles3[2] = GetMsg(MName);
  columnTitles3[3] = GetMsg(MSectorSize);

  static const wchar_t*     columnTitles4[4];
  columnTitles4[0] = GetMsg(MName);
  columnTitles4[1] = GetMsg(MSize);
  columnTitles4[2] = GetMsg(MStart);
  columnTitles4[3] = GetMsg(MSectorSize);

  static const wchar_t*     columnTitles5[2];
  columnTitles5[0] = GetMsg(MName);
  columnTitles5[1] = GetMsg(MFormat);

  static const wchar_t*     columnTitles6[2];
  columnTitles6[0] = GetMsg(MName);
  columnTitles6[1] = GetMsg(MComment);
  
  // C0 - trdos name
  // C1 - trdos 'start'
  // C2 -   
  // C3 -  
  // C4 -   
  
  mode[3].ColumnTypes=L"N,C2,N,C2";
  mode[3].ColumnWidths=L"0,3,0,3";
  mode[3].ColumnTitles=columnTitles3;
  mode[3].StatusColumnTypes=L"C0,S,C2";
  mode[3].StatusColumnWidths=L"0,5,3";
  mode[3].Flags = PMFLAGS_ALIGNEXTENSIONS;
  
  mode[4].ColumnTypes=L"C0,S,C1,C2";
  mode[4].ColumnWidths=L"0,5,5,3";
  mode[4].StatusColumnTypes=L"N,S,C2";
  mode[4].StatusColumnWidths=L"0,5,3";
  mode[4].ColumnTitles=columnTitles4;
  
  mode[5].ColumnTypes=L"C0,C3";
  mode[5].ColumnWidths=L"12,0";
  mode[5].StatusColumnTypes=L"N,S,C2";
  mode[5].StatusColumnWidths=L"0,5,3";
  mode[5].ColumnTitles=columnTitles5;

  mode[6].ColumnTypes=L"C0,C4";
  mode[6].ColumnWidths=L"12,0";
  mode[6].StatusColumnTypes=L"N,S,C2";
  mode[6].StatusColumnWidths=L"0,5,3";
  mode[6].ColumnTitles=columnTitles6;
  
  Info->PanelModesArray = mode;
  Info->PanelModesNumber = std::size(mode);
  Info->StartPanelMode = L'0' + op.defaultPanelMode;
}

int Manager::getFindData(PluginPanelItem **pPanelItem, size_t *pNoItems, int opMode)
{
  if(!readInfo()) return FALSE;
  
  PluginPanelItem *item = new PluginPanelItem[noFiles]{};
  *pNoItems   = noFiles;
  *pPanelItem = item;
  
  if(noFiles == 0) return TRUE;
  
  //      trdos' 
  PanelInfo panel = { sizeof(PanelInfo) };
  startupInfo.PanelControl(this, FCTL_GETPANELINFO, 0, &panel);
  int nameColumnWidth = 12;
  
  if(panel.ViewMode == 4)
  {
    nameColumnWidth = panel.PanelRect.right - panel.PanelRect.left - 2 - 15;
  }
  if(nameColumnWidth < 12) nameColumnWidth = 12;

  wchar_t* name = new wchar_t[nameColumnWidth+1];

  for(int fNum = 0; fNum < noFiles; ++fNum)
  {
    item[fNum].FileAttributes = files[fNum].name[0] == 0x01 ? FILE_ATTRIBUTE_HIDDEN : FILE_ATTRIBUTE_NORMAL;
    item[fNum].FileName = pcFiles[fNum].name;
    item[fNum].FileSize = files[fNum].size;

    item[fNum].CustomColumnNumber = 5;
    wchar_t **CustomColumnData = new LPWSTR[item[fNum].CustomColumnNumber];
    item[fNum].CustomColumnData = CustomColumnData;

    CustomColumnData[0] = new wchar_t[nameColumnWidth+1];

    wmemset(name, L' ', nameColumnWidth + 1);

    makeTrDosName(name, files[fNum], op.showExt, nameColumnWidth);
    lstrcpyW(CustomColumnData[0], name);

    CustomColumnData[1] = new wchar_t[6];
    wsprintfW(CustomColumnData[1], L"%5d", files[fNum].start);

    CustomColumnData[2] = new wchar_t[4];
    wsprintfW(CustomColumnData[2],
             L"%3d",
             files[fNum].noSecs);

    const wchar_t* description = detector->description(pcFiles[fNum].type);
    if(description)
    {
      CustomColumnData[3] = new wchar_t[lstrlenW(description)+1];
      lstrcpyW(CustomColumnData[3], description);
    }
    else
    {
      CustomColumnData[3] = new wchar_t[1];
      CustomColumnData[3][0] = 0;
    }

    const wchar_t* comment = pcFiles[fNum].comment;
    if(comment[0])
    {
      CustomColumnData[4] = new wchar_t[lstrlenW(comment)+1];
      lstrcpyW(CustomColumnData[4], comment);
    }
    else
    {
      CustomColumnData[4] = new wchar_t[1];
      CustomColumnData[4][0] = 0;
    }
  }
  delete[] name;
  return TRUE;
}

void Manager::freeFindData(PluginPanelItem *panelItem, int noItems)
{
  for (int i = 0; i < noItems; i++)
  {
    delete[] panelItem[i].CustomColumnData[0];
    delete[] panelItem[i].CustomColumnData[1];
    delete[] panelItem[i].CustomColumnData[2];
    delete[] panelItem[i].CustomColumnData[3];
    delete[] panelItem[i].CustomColumnData[4];
    delete[] panelItem[i].CustomColumnData;
  }
  delete[] panelItem;
}
