#ifndef MANAGER_HPP
#define MANAGER_HPP

#include <windows.h>
#include "plugin.hpp"
#include "fmtReader.hpp"
#include "types.hpp"

class Manager
{
  public:
    Manager(const wchar_t* fileName, const FmtPlugin *fmtPlugin);
    ~Manager();
    void GetOpenPanelInfo(struct OpenPanelInfo *Info);

    int  getFindData      (PluginPanelItem **pPanelItem, size_t *pNoItems, int opMode);
    void freeFindData     (PluginPanelItem *panelItem,   int  noItems);
    int  setDirectory     (const wchar_t* dirName, OPERATION_MODES opMode);
    int  makeDirectory    (const wchar_t *dirName, OPERATION_MODES opMode);
    int  deleteFiles      (PluginPanelItem *panelItem, int noItems, int opMode);
    int  getFiles         (PluginPanelItem *panelItem, size_t noItems, BOOL move, const wchar_t *destPath, OPERATION_MODES opMode);
    int  putFiles         (PluginPanelItem *panelItem, int noItems, int move, int opMode);
    int  processKey       (const INPUT_RECORD *Rec);
    int  diskMenu         ();
    int  processHostFile  (PluginPanelItem* panelItem, int noItems, int opMode);
  private:
    wchar_t         hostFileName[260];
    wchar_t         curFolder   [260];
    int             curFolderNum;
    WIN32_FIND_DATAW lastModifed;
    const FmtPlugin       *fmt;
    int             noFiles, noDelFiles;
    HANDLE          img;

    FileHdr         files  [142];
    ExtFileHdr      pcFiles[142];
    DiskHdr         diskInfo;

    BYTE            fileMap  [142];
    BYTE            folderMap[127];
    char            folders  [127][11];
    wchar_t         pcFolders[127][15];
    int             noFolders, noDelFolders;

    BYTE            zeroTrk[16*sectorSize];
    
    WORD            dsCRC;
    bool            dsOk;

    void getBytes     (void* to, int size);
    bool readInfo     (void);
    void move         (void);
    void writeInfo    (void);
    bool openHostFile (void);
    void closeHostFile(void);
    void makePCNames  (void);
    WORD calcDScrc    (void);
    bool checkEntry   (char num, bool del, bool *checked);
    bool checkDS      (void);
    bool checkDisk    (void);
    
    ExitCode markFolder(int fNum);

    ExitCode getFolder (int fNum, bool isMove, const wchar_t* to);
    ExitCode getFile   (int fNum, bool isMove, const wchar_t* to);

    bool isDiskFull   (int noAddFiles, int noSecs);
    bool makeFolder   (const wchar_t* folderName, BYTE folderNum);
   
    ExitCode writeFile(const SCLHdr& h, HANDLE file, BYTE folderNum);
    ExitCode putFile  (const wchar_t *fileName,   BYTE folderNum, bool move);
    ExitCode putFolder(const wchar_t* folderName, BYTE folderNum, bool move);

    bool read         (int trk, int sec, BYTE* buf);
    bool write        (int trk, int sec, BYTE* buf);
};

#endif
